/* Distributed Checksum Clearinghouse
 *
 * dump list of dccd clients
 *
 * Copyright (c) 2008 by Rhyolite Software, LLC
 *
 * This agreement is not applicable to any entity which sells anti-spam
 * solutions to others or provides an anti-spam solution as part of a
 * security solution sold to other entities, or to a private network
 * which employs the DCC or uses data provided by operation of the DCC
 * but does not provide corresponding data to other users.
 *
 * Permission to use, copy, modify, and distribute this software without
 * changes for any purpose with or without fee is hereby granted, provided
 * that the above copyright notice and this permission notice appear in all
 * copies and any distributed versions or copies are either unchanged
 * or not called anything similar to "DCC" or "Distributed Checksum
 * Clearinghouse".
 *
 * Parties not eligible to receive a license under this agreement can
 * obtain a commercial license to use DCC by contacting Rhyolite Software
 * at sales@rhyolite.com.
 *
 * A commercial license would be for Distributed Checksum and Reputation
 * Clearinghouse software.  That software includes additional features.  This
 * free license for Distributed ChecksumClearinghouse Software does not in any
 * way grant permision to use Distributed Checksum and Reputation Clearinghouse
 * software
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND RHYOLITE SOFTWARE, LLC DISCLAIMS ALL
 * WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL RHYOLITE SOFTWARE, LLC
 * BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
 * SOFTWARE.
 *
 * Rhyolite Software DCC 1.3.90-1.9 $Revision$
 */

#include "dccd_defs.h"

#define PROGNAME "dump-clients: "


static const char *fname;
static FILE *f;


static void NRATTRIB
usage(void)
{
	fprintf(stderr, "usage: [-anq] ifile1 ifile2 ...\n");
	exit(1);
}



static void PATTRIB(1,2)
error_msg(const char *p, ...)
{
	va_list args;

	fflush(stderr);

	va_start(args, p);
	vfprintf(stderr, p, args);
	va_end(args);
	putchar('\n');
}



static u_char
clients_read(off_t pos, void *buf, int buf_len)
{
	int i;

	if (fseeko(f, pos, 0)) {
		error_msg("fseeko(%s,%d): %s", fname, (int)pos, ERROR_STR());
		fclose(f);
		return 0;
	}

	i = fread(buf, buf_len, 1, f);
	if (i == 1)
		return 1;

	if (feof(f))
		return 0;

	error_msg("fread(%s): %s", fname, ERROR_STR());
	fclose(f);
	return 0;
}



int NRATTRIB
main(int argc, char **argv)
{
	u_char nonames, quiet, avg;
	struct stat sb;
	CLIENTS_HEADER header;
	RL_DATA data;
	struct {
	    u_int   hosts;
	    u_int   requests;
	    u_int   nops;
	}  versions[DCC_PKT_VERSION_MAX+1];
	struct tm tm;
	char date_buf[40];
	int inflate;
	double delay_us;
	DCC_SOCKU su;
	char name[DCC_MAXDOMAINLEN];
	char sustr[DCC_SU2STR_SIZE];
	char avg_buf[1+6+1+1+6+1+1];
	off_t pos;
	int fno, total, active, i;

	avg = 0;
	nonames = 0;
	quiet = 0;

	while ((i = getopt(argc, argv, "anq")) != -1) {
		switch (i) {
		case 'a':
			++avg;
			break;
		case 'n':
			nonames = 1;
			break;
		case 'q':
			quiet = 1;
			break;
		default:
			usage();
		}
	}
	argc -= optind;
	argv += optind;
	if (argc < 1)
		usage();

	for (fno = 0; fno < argc; ++fno) {
		fname = argv[fno];

		f = fopen(fname, "r");
		if (!f) {
			error_msg(PROGNAME"stat(%s): %s",
				  fname, ERROR_STR());
			exit(1);
		}

		if (0 > fstat(fileno(f), &sb)) {
			error_msg("stat(%s): %s", fname, ERROR_STR());
			exit(1);
		}
		pos = sb.st_size;
		if (pos < ISZ(header)
		    || ((pos - ISZ(header)) % ISZ(RL_DATA)) != 0) {
			error_msg("%s has impossible size %d", fname, (int)pos);
			exit(1);
		}

		if (!clients_read(0, &header, sizeof(header)))
			exit(1);
		if (strcmp(header.magic, CLIENTS_MAGIC(0))
		    && strcmp(header.magic, CLIENTS_MAGIC(1))) {
			error_msg("unrecognized magic in %s", fname);
			exit(1);
		}
		if (header.hash_len > RL_MIN_MAX_MAX) {
			error_msg("unrecognized hash_len=%d in %s",
				  header.hash_len, fname);
			exit(1);
		}
		strftime(date_buf, sizeof(date_buf), "%m/%d %X",
			 gmtime_r(&header.cleared, &tm));
		printf("%s\n    cleared %s  anon delay=%d  inflate=%d\n",
		       fname, date_buf,
		       header.anon_delay_us, header.anon_delay_inflate);

		total = 0;
		active = 0;
		memset(versions, 0, sizeof(versions));
		while ((pos -= ISZ(RL_DATA)) > ISZ(header)) {
		     if (!clients_read(pos, &data, sizeof(data)))
				break;

			++total;
			if (data.requests)
				++active;

			if (data.pkt_vers < DIM(versions))
				i = data.pkt_vers;
			else
				i = 0;
			++versions[i].hosts;
			versions[i].requests += data.requests;
			versions[i].nops += data.nops;

			if (quiet)
				continue;

			snprintf(avg_buf, sizeof(avg_buf), " (%d,%d)",
				 data.requests_cur_avg,
				 RL_REQUESTS_NEXT_AVG(&data));

			if (data.requests != 0 || data.nops != 0) {
				printf("%7d %6d", data.requests, data.nops);
			} else {
				printf("%7s %6s", "-", "-");
			}

			if (avg)
				printf("%16s", avg_buf);

			strftime(date_buf, sizeof(date_buf), "%m/%d %X",
				 gmtime_r(&data.last_used, &tm));
			printf(" %s %6d", date_buf, data.clnt_id);

			if (data.pkt_vers != 0)
				printf(" %d", data.pkt_vers);
			else
				fputs(" ?", stdout);

			if (data.flags & RL_FG_ADDR_BL) {
				fputs(" A BL ", stdout);
			} else if (data.flags & RL_FG_ID_BL) {
				fputs(" ID BL", stdout);
			} else if (data.clnt_id == DCC_ID_ANON
				   && header.anon_delay_inflate != 0) {
				inflate = 1 + (RL_REQUESTS_AVG(&data)
					       / header.anon_delay_inflate);
				if (inflate > (DCC_ANON_DELAY_MAX
					       / header.anon_delay_us)) {
					delay_us = DCC_ANON_DELAY_MAX;
				} else {
					delay_us = inflate*header.anon_delay_us;
				}
				printf(" %5.3f", delay_us/DCC_US);
			} else {
				fputs("      ", stdout);
			}

			dcc_mk_su(&su, AF_INET6, &data.clnt_addr, 0);
			if (nonames) {
				printf(" %s\n",
				       dcc_su2str2(sustr, sizeof(sustr), &su));
			} else {
				printf(" %-16s %s\n",
				       dcc_su2str2(sustr, sizeof(sustr), &su),
				       dcc_su2name(name, sizeof(name), &su));
			}
		}

		fclose(f);

		printf("%d of %d blocks active\n\n", active, total);

		fputs("version  hosts requests     nops\n", stdout);
		for (i = 0; i < DIM(versions); ++i) {
			if (versions[i].hosts == 0)
				continue;
			printf("%6d %6d %8d %8d\n",
			       i, versions[i].hosts,
			       versions[i].requests, versions[i].nops);
		}
	}

	exit(0);
}
