/* Distributed Checksum Clearinghouse
 *
 * server rate limiting
 *
 * Copyright (c) 2008 by Rhyolite Software, LLC
 *
 * This agreement is not applicable to any entity which sells anti-spam
 * solutions to others or provides an anti-spam solution as part of a
 * security solution sold to other entities, or to a private network
 * which employs the DCC or uses data provided by operation of the DCC
 * but does not provide corresponding data to other users.
 *
 * Permission to use, copy, modify, and distribute this software without
 * changes for any purpose with or without fee is hereby granted, provided
 * that the above copyright notice and this permission notice appear in all
 * copies and any distributed versions or copies are either unchanged
 * or not called anything similar to "DCC" or "Distributed Checksum
 * Clearinghouse".
 *
 * Parties not eligible to receive a license under this agreement can
 * obtain a commercial license to use DCC by contacting Rhyolite Software
 * at sales@rhyolite.com.
 *
 * A commercial license would be for Distributed Checksum and Reputation
 * Clearinghouse software.  That software includes additional features.  This
 * free license for Distributed ChecksumClearinghouse Software does not in any
 * way grant permision to use Distributed Checksum and Reputation Clearinghouse
 * software
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND RHYOLITE SOFTWARE, LLC DISCLAIMS ALL
 * WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL RHYOLITE SOFTWARE, LLC
 * BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
 * SOFTWARE.
 *
 * Rhyolite Software DCC 1.3.90-1.124 $Revision$
 */

#include "dccd_defs.h"


RL_RATE rl_sub_rate;
RL_RATE rl_anon_rate;
RL_RATE rl_all_anon_rate;
RL_RATE rl_bugs_rate;
#ifdef RL_MIN_MAX
static int rl_min_max = RL_MIN_MAX;
#else
static int rl_min_max = 0;
#endif

static RL rl_all_anon;			/* limit all or anonymous clients */

static RL *rl_newest, *rl_oldest;
static RL **rl_hash;
static int rl_hash_len;
static u_char rl_too_many;

time_t clients_cleared;

time_t clients_saved;

typedef struct ip_bl {
    struct ip_bl *fwd;
    struct in6_addr addr;
    struct in6_addr mask;
    u_char  flags;			/* subset of RL_FG_* */
#    define  IP_BL_ADDR_BL  RL_FG_ADDR_BL
#    define  IP_BL_TRACE    RL_FG_TRACE
} IP_BL;
static IP_BL *ip_bl;

static u_char rl_get(RL **, DCC_CLNT_ID, const struct in6_addr *);



/* See if an IP address is evil
 *	The blacklist is a simple linear list.  Entries from the file
 *	are added to the front of the file, so that the last matching
 *	entry determines the result.
 *  This should be sped up if there are ever more than a very few entries */
u_char					/* 1=blacklisted */
ck_ip_bl(RL **rlp, DCC_CLNT_ID clnt_id, const struct in6_addr *pap)
{
	const IP_BL *bl;
	RL *rl;

	for (bl = ip_bl; bl; bl = bl->fwd) {
		if (DCC_IN_BLOCK(*pap, bl->addr, bl->mask)) {
			/* all blacklisted addresses including flooding peers
			 * should have rate limit blocks so that they are
			 * noticed */
			if (rlp && *rlp)
				rl = *rlp;
			else
				rl_get(&rl, clnt_id, pap);
			rl->d.flags &= ~(RL_FG_ADDR_BL | RL_FG_TRACE);
			rl->d.flags |= (RL_FG_CK_BL | bl->flags);
			if (rlp && !*rlp)
				*rlp = rl;
			return (rl->d.flags & RL_FG_ADDR_BL);
		}
	}
	return 0;
}



static void
clear_bl(void)
{
	IP_BL *bl;
	RL *rl;

	while ((bl = ip_bl) != 0) {
		ip_bl = ip_bl->fwd;
		dcc_free(bl);
	}

	for (rl = rl_newest; rl != 0; rl = rl->older) {
		rl->d.flags &= ~RL_FG_CK_BL;
	}
}



void
check_blacklist_file(void)
{
#define BL_NM "blacklist"
	DCC_FNM_LNO_BUF fnm_buf;
	static time_t prev_mtime, next_msg;
#define BL_RECHECK (2*3600)
	static int serrno;
	struct stat sb;
	FILE *f;
	char buf[120];
	IP_BL *bl;
	struct in6_addr addr, mask;
	u_char flags;
	char *p;
	int lno, entries, i;

	/* see if the file has changed */
	if (0 > stat(BL_NM, &sb)) {
		if (errno != ENOENT) {
			prev_mtime = 1;
			if (serrno != errno
			    || DB_IS_TIME(next_msg, BL_RECHECK)) {
				serrno = errno;
				next_msg = db_time.tv_sec + BL_RECHECK;
				dcc_error_msg("stat(%s): %s",
					      DB_NM2PATH_ERR(BL_NM),
					      ERROR_STR());
			}

		} else if (prev_mtime != 0) {
			prev_mtime = 0;
			next_msg = 0;
			dcc_error_msg("%s disappeared", DB_NM2PATH_ERR(BL_NM));
		}
		clear_bl();
		return;
	}
	if (prev_mtime == sb.st_mtime
	    && (next_msg == 0 || !DB_IS_TIME(next_msg, BL_RECHECK)))
		return;

	/* the file has changed, so parse it */
	clear_bl();
	prev_mtime = 1;
	next_msg = 0;
	f = fopen(BL_NM, "r");
	if (!f) {
		if (serrno != errno) {
			serrno = errno;
			next_msg = db_time.tv_sec + BL_RECHECK;
			dcc_error_msg("fopen(%s): %s",
				      DB_NM2PATH_ERR(BL_NM), ERROR_STR());
		}
		return;
	}
	if (0 > fstat(fileno(f), &sb)) {
		if (serrno != errno) {
			serrno = errno;
			next_msg = db_time.tv_sec + BL_RECHECK;
			dcc_error_msg("fstat(%s): %s",
				      DB_NM2PATH_ERR(BL_NM), ERROR_STR());
		}
		return;
	}
	prev_mtime = sb.st_mtime;

	entries = 0;
	for (lno = 1; ; ++lno) {
		if (!fgets(buf, sizeof(buf), f)) {
			if (ferror(f) && serrno != errno) {
				serrno = errno;
				next_msg = db_time.tv_sec + BL_RECHECK;
				dcc_error_msg("fgets(%s): %s",
					      DB_NM2PATH_ERR(BL_NM),
					      ERROR_STR());
			}
			break;
		}
		/* reject lines that are too long */
		i = strlen(buf);
		if (buf[i-1] != '\n') {
			next_msg = db_time.tv_sec + BL_RECHECK;
			dcc_error_msg("syntax error%s",
				      fnm_lno(fnm_buf,
					      DB_NM2PATH_ERR(BL_NM), lno));
			break;
		}

		/* ignore leading blanks, comments, and blank lines */
		p = strchr(buf, '#');
		if (p)
			*p = '\0';
		p = buf;
		p += strspn(p, DCC_WHITESPACE);
		if (*p == '\0')
			continue;

		flags = IP_BL_ADDR_BL;
		for (;;) {
			i = strspn(p, " \t");
			if (i != 0) {
				p += i;
				continue;
			}
			if (!CLITCMP(p, "trace")) {
				p += LITZ("trace");
				if (*p == ',')
					++p;
				flags |= IP_BL_TRACE;
				continue;
			}
			if (!CLITCMP(p, "ok")) {
				p += LITZ("ok");
				if (*p == ',')
					++p;
				flags &= ~IP_BL_ADDR_BL;
				continue;
			}
			if (!CLITCMP(p, "bad")) {
				p += LITZ("bad");
				if (*p == ',')
					++p;
				flags |= IP_BL_ADDR_BL;
				continue;
			}
			break;
		}

		i = dcc_str2cidr(dcc_emsg, &addr, &mask, p, BL_NM, lno);
		if (i <= 0) {
			if (i < 0)
				dcc_error_msg("%s", dcc_emsg);
			else
				dcc_error_msg("syntax error%s",
					      fnm_lno(fnm_buf,
						      DB_NM2PATH_ERR(BL_NM),
						      lno));
			next_msg = db_time.tv_sec + BL_RECHECK;
			continue;
		}

		bl = dcc_malloc(sizeof(*bl));
		bl->addr = addr;
		bl->mask = mask;
		bl->flags = flags;
		bl->fwd = ip_bl;
		ip_bl = bl;
		if (++entries > 100) {
			dcc_error_msg("too many entries in %s",
				      DB_NM2PATH_ERR(BL_NM));
			next_msg = db_time.tv_sec + BL_RECHECK;
			break;
		}
	}
	fclose(f);

	if (entries)
		dcc_trace_msg("read %d entries from %s",
			      entries, DB_NM2PATH_ERR(BL_NM));

#undef BL_NM
#undef BL_RECHECK
}



static RL **
rl_hash_fnc(DCC_CLNT_ID	id, const struct in6_addr *addr)
{
	u_int sum;

	sum = id;
	sum += addr->s6_addr32[0];
	sum += addr->s6_addr32[1];
	sum += addr->s6_addr32[2];
	sum += addr->s6_addr32[3];
	return &rl_hash[mhash(sum, rl_hash_len)];
}



static inline void
rl_hash_link(RL *rl, RL **rlh)
{
	RL *rl2;

	rl->hash = rlh;
	rl2 = *rlh;
	rl->fwd = rl2;
	if (rl2)
		rl2->bak = rl;
	*rlh = rl;
}



static int
set_rl_min_max(void)
{
	int i;


	if (rl_min_max == 0) {
		/* allow RL_MIN_MAX_DEF for 512 MByte
		 *	RL_MIN_MAX_DEF*2 for 1 GByte
		 *	...
		 *	RL_MIN_MAX_DEF*16 for 8 GByte
		 */
		i = (db_max_rss + 512*1024*1024-1) / (512*1024*1024);
		if (i <= 0)
			i = 1;
		rl_min_max = RL_MIN_MAX_DEF * i;
	}
	if (rl_min_max > RL_MIN_MAX_MAX)
		rl_min_max = RL_MIN_MAX_MAX;

	return rl_min_max;
}



static void
rl_expand(int new_len)
{
	RL *rl;
	int j;

	/* create a bunch of rate limit blocks */
	if (new_len < 64)
		new_len = 64;

	if (rl_hash_len+new_len > set_rl_min_max()) {
		if (rl_hash_len > rl_min_max)
			dcc_logbad(EX_SOFTWARE, "impossible # of RL blocks");
		new_len = rl_min_max - rl_hash_len;
	}

	if (rl_hash_len != 0)
		dcc_trace_msg("increase from %d to %d RL blocks",
			      rl_hash_len, rl_hash_len+new_len);

	rl = dcc_malloc(new_len*sizeof(*rl));
	if (!rl)
		dcc_logbad(EX_OSERR, "malloc(%d RL's) failed", new_len);
	memset(rl, 0, new_len*sizeof(*rl));
	j = 0;
	if (!rl_oldest) {
		rl_oldest = rl;
		rl_newest = rl;
		++rl;
		++j;
	}
	while (j < new_len) {	/* make the new blocks oldest */
		rl_oldest->older = rl;
		rl->newer = rl_oldest;
		rl_oldest = rl;
		 ++rl;
		 ++j;
	}

	/* Rebuild and expand the hash table.
	 *	The hash table is an array of pointers to rate limit blocks. */
	if (rl_hash)
		dcc_free(rl_hash);
	rl_hash_len += new_len;
	rl_hash = dcc_malloc(rl_hash_len*sizeof(*rl_hash));
	memset(rl_hash, 0, rl_hash_len*sizeof(RL *));

	/* copy the active blocks to the new hash table */
	for (rl = rl_newest; rl; rl = rl->older) {
		if (!rl->hash)
			continue;
		rl->bak = 0;
		rl_hash_link(rl, rl_hash_fnc(rl->d.clnt_id, &rl->d.clnt_addr));
	}
}



/* age a rate limit */
static void
rl_age(RL *rl, const RL_RATE *credits)
{
	time_t secs;

	secs = (db_time.tv_sec - rl->d.last_used);

	/* only prevent overflow if no time has passed */
	if (secs == 0) {
		if (rl->d.request_credits < credits->lo)
			rl->d.request_credits = credits->lo;
		if (rl->d.bug_credits < rl_bugs_rate.lo)
			rl->d.bug_credits = rl_bugs_rate.lo;
	}

	rl->d.last_used = db_time.tv_sec;

	/* reset idle counters */
	if (secs >= RL_AVG_SECS || secs < 0) {
		rl->d.bug_credits = rl_bugs_rate.hi;
		rl->d.request_credits = credits->hi;
		return;
	}

	rl->d.request_credits += secs*credits->sec;
	if (rl->d.request_credits > credits->hi)
		rl->d.request_credits = credits->hi;

	rl->d.bug_credits += secs*rl_bugs_rate.sec;
	if (rl->d.bug_credits > rl_bugs_rate.hi)
		rl->d.bug_credits = rl_bugs_rate.hi;
}



/* age and increment a rate limit */
void
rl_inc(RL *rl, const RL_RATE *credits)
{
	rl_age(rl, credits);
	++rl->d.requests;		/* increase total count of requests */
	rl->d.request_credits -= RL_SCALE;  /* charge for this request */
}



static void
rl_unref(RL *rl)
{
	if (rl->newer) {
		rl->newer->older = rl->older;
	} else if (rl_newest == rl) {
		rl_newest = rl->older;
	}

	if (rl->older) {
		rl->older->newer = rl->newer;
	} else if (rl_oldest == rl) {
		rl_oldest = rl->newer;
	}

	rl->newer = 0;
	rl->older = 0;
}



static void
rl_ref(RL *rl)
{
	if (rl_newest == rl)
		return;

	rl_unref(rl);
	rl->older = rl_newest;
	rl_newest->newer = rl;
	rl_newest = rl;
}



/* get a free rate limit block, recycling the oldest block if necessary */
static RL *
rl_get_free(void)
{
	RL *rl, *rl2;
	time_t stale;

	for (rl = rl_oldest; rl != 0; rl = rl->newer) {
		if (rl->ref_cnt)
			continue;

		/* Found oldest free block */
		if (rl->d.last_used != 0) {
			/* oldest free block has been used and so we
			 * must recycle something */
			stale = db_time.tv_sec;
			/* keep many blocks until we get enough to
			 * worry about a denial of service attack */
			if (rl_hash_len < set_rl_min_max())
				stale -= CLIENTS_SAVE_AGE;
			else
				stale -= RL_LIFE_SECS;
			/* make more if the oldest is new
			 * and we don't have too many */
			if (rl->d.last_used >= stale
			    && rl_hash_len < rl_min_max)
				return 0;

			/* Notice if we are about to recycle a block that is
			 * not obsolete.
			 * Try to find an old, little used block, so that we
			 * do not recycle a block used by a busy client that
			 * has only paused */
			if (rl->d.last_used > clients_cleared) {
				rl_too_many = 1;
				stale = db_time.tv_sec - CLIENTS_AGE/2;
				for (rl2 = rl;
				     rl2 && rl2->d.last_used <= stale;
				     rl2 = rl2->newer) {
					/* avoid forgeting bad guys */
					if (rl->d.flags & (RL_FG_ADDR_BL
							| RL_FG_ID_BL))
					    continue;
					/* take first tiny client found */
					if (rl2->d.requests < 100
					    && RL_REQUESTS_AVG(&rl2->d) < 100) {
					    rl = rl2;
					    break;
					}
				}
			}
		}

		/* recycle a block */
		if (rl->fwd)
			rl->fwd->bak = rl->bak;
		if (rl->bak)
			rl->bak->fwd = rl->fwd;
		else if (rl->hash)
			*rl->hash = rl->fwd;
		rl_unref(rl);
		memset(rl, 0, sizeof(*rl));

		rl->d.requests_avg_date = db_time.tv_sec + RL_AVG_DAY;
		rl_ref(rl);
		return rl;
	}

	/* there are no free blocks that are old enough to recycle */
	return 0;
}



/* update the number of requests made yesterday if it is time */
static void
rl_avg_requests_age(RL *rl,
		    u_char force)	/* sync to daily `cdcc "stats clear"` */
{
	if (rl->d.requests_avg_date <= db_time.tv_sec
	    || rl->d.requests_avg_date + RL_AVG_DAY < db_time.tv_sec
	    || force) {
		rl->d.requests_cur_avg = (rl->d.requests_next_avg
					  + rl->d.requests);
		rl->d.nops_cur_avg = rl->d.nops_next_avg + rl->d.nops;

		/* these biases effectively zero tomorrow's averages */
		rl->d.requests_next_avg = -rl->d.requests;
		rl->d.nops_next_avg = -rl->d.nops;

		rl->d.requests_avg_date = db_time.tv_sec + RL_AVG_DAY;
	}
}



/* get a rate limit block based on the IP address of the sender */
static u_char				/* 1=existing block */
rl_get(RL **rlp, DCC_CLNT_ID clnt_id, const struct in6_addr *cap)
{
	RL *rl, **rlh;

	if (!rl_hash_len)
		rl_expand(queue_max);

	rlh = rl_hash_fnc(clnt_id, cap);
	for (rl = *rlh; rl; rl = rl->fwd) {
		if (rl->d.clnt_id != clnt_id
		    || memcmp(&rl->d.clnt_addr, cap, sizeof(rl->d.clnt_addr)))
			continue;
		rl_ref(rl);		/* found it, so make it newest */
		rl_avg_requests_age(rl, 0);
		*rlp = rl;
		return 1;
	}

	rl = rl_get_free();
	if (!rl) {
		/* when we are out of rate limiting blocks, make more */
		rl_expand(queue_max);
		/* which makes a new rate limit hash table which makes our
		 * pointer into the old hash table bogus */
		rl = rl_get_free();
		if (!rl)
			dcc_logbad(EX_SOFTWARE, "no available RL blocks");
		rlh = rl_hash_fnc(clnt_id, cap);
	}
	rl_hash_link(rl, rlh);
	rl->d.clnt_addr = *cap;
	rl->d.clnt_id = clnt_id;
	*rlp = rl;
	return 0;
}



/* get a rate limit block for a queue entry */
static RL *
rl_get_q(QUEUE *q)
{
	RL *rl;
	struct in6_addr clnt_addr;
	const struct in6_addr *cap;

	if (q->clnt_su.sa.sa_family == AF_INET6) {
		cap = &q->clnt_su.ipv6.sin6_addr;
	} else {
		dcc_ipv4toipv6(&clnt_addr, q->clnt_su.ipv4.sin_addr);
		cap = &clnt_addr;
	}

	rl_get(&rl, q->clnt_id, cap);

	q->rl = rl;
	++rl->ref_cnt;

	/* See if this client IP address is blacklisted, if we have not
	 * checked it recently. */
	if (!(rl->d.flags & RL_FG_CK_BL)) {
		rl->d.flags |= RL_FG_CK_BL;
		rl->d.flags &= ~(RL_FG_ADDR_BL | RL_FG_TRACE);
		ck_ip_bl(&rl, q->clnt_id, &rl->d.clnt_addr);
	}

	return rl;
}



static u_char
clients_write(FILE *f, const void *buf, int buf_len)
{
	if (1 == fwrite(buf, buf_len, 1, f))
		return 1;
	dcc_error_msg("fwrite(%s): %s",
		      DB_NM2PATH_ERR(CLIENTS_NM()), ERROR_STR());
	fclose(f);
	return 0;
}



/* dump the rate limit blocks into a file */
void
clients_save(void)
{
	int fd;
	FILE *f;
	CLIENTS_HEADER header;
	RL *rl;
	time_t stale;

	clients_saved = db_time.tv_sec + CLIENTS_SAVE_SECS;

	stale = db_time.tv_sec - CLIENTS_SAVE_AGE;

	/* prevent evil games with symbolic links */
	unlink(CLIENTS_NM());
	fd = open(CLIENTS_NM(), O_WRONLY|O_CREAT|O_EXCL, 0644);
	if (fd < 0) {
		dcc_error_msg("open(%s): %s",
			      DB_NM2PATH_ERR(CLIENTS_NM()), ERROR_STR());
		return;
	}
	f = fdopen(fd, "w");

	memset(&header, 0, sizeof(header));
	BUFCPY(header.magic, CLIENTS_MAGIC(grey_on));
	header.anon_delay_us = anon_delay_us;
	header.anon_delay_inflate = anon_delay_inflate;
	header.cleared = clients_cleared;
	if (header.cleared < stale)
		header.cleared = stale;

	for (rl = rl_oldest; rl != 0; rl = rl->newer) {
		if (rl->d.last_used >= stale)
			++header.hash_len;
	}
	if (!clients_write(f, &header, sizeof(header)))
		return;

	for (rl = rl_oldest; rl != 0; rl = rl->newer) {
		if (rl->d.last_used < stale)
			continue;
		if (!clients_write(f, &rl->d, sizeof(rl->d)))
			return;
	}
	fclose(f);
}



static u_char
clients_read(FILE *f, void *buf, int buf_len)
{
	int i;

	i = fread(buf, buf_len, 1, f);
	if (i == 1)
		return 1;

	if (feof(f))
		return 0;

	dcc_error_msg("fread(%s): %s",
		      DB_NM2PATH_ERR(CLIENTS_NM()), ERROR_STR());
	fclose(f);
	return 0;
}



/* load the rate limit blocks from a previous instance */
void
clients_load(void)
{
	struct stat sb;
	int fd;
	FILE *f;
	CLIENTS_HEADER header;
	RL_DATA data;
	RL *rl;
	int i, max, total, used;
	time_t stale;
#	define BAD_FILE() ((0 > rename(CLIENTS_NM(), BAD_CLIENTS_NM()))	    \
			   ? unlink(CLIENTS_NM()) : 0)

	clients_cleared = db_time.tv_sec;

	fd = open(CLIENTS_NM(), O_RDONLY, 0);
	if (fd < 0) {
		if (errno != ENOENT)
			dcc_error_msg("open(%s): %s",
				      DB_NM2PATH_ERR(CLIENTS_NM()),
				      ERROR_STR());
		else if (db_debug)
			dcc_trace_msg("open(%s): %s",
				      DB_NM2PATH_ERR(CLIENTS_NM()),
				      ERROR_STR());
		BAD_FILE();
		return;
	}
	f = fdopen(fd, "r");
	if (0 > fstat(fd, &sb)) {
		dcc_error_msg("stat(%s): %s",
			      DB_NM2PATH_ERR(CLIENTS_NM()), ERROR_STR());
		BAD_FILE();
		fclose(f);
		return;
	}
	i = sb.st_size;
	if (i < ISZ(header)
	    || ((i - ISZ(header)) % ISZ(RL_DATA)) != 0) {
		dcc_error_msg("%s has impossible size %d",
			      DB_NM2PATH_ERR(CLIENTS_NM()), i);
		BAD_FILE();
		fclose(f);
		return;
	}

	if (!clients_read(f, &header, sizeof(header)))
		return;
	if (strcmp(header.magic, CLIENTS_MAGIC(grey_on))) {
		if (strncmp(header.magic, CLIENTS_MAGIC_BASE(grey_on),
			    strlen(CLIENTS_MAGIC_BASE(grey_on)))) {
			dcc_error_msg("unrecognized magic in %s",
				      DB_NM2PATH_ERR(CLIENTS_NM()));
			BAD_FILE();
		}
		fclose(f);
		return;
	}
	if (header.hash_len > RL_MIN_MAX_MAX) {
		dcc_trace_msg("unrecognized hash_len=%d in %s",
			      header.hash_len, DB_NM2PATH_ERR(CLIENTS_NM()));
		fclose(f);
		return;
	}
	if (header.cleared > db_time.tv_sec) {
		dcc_trace_msg("bad time %d in %s",
			      (int)header.cleared,
			      DB_NM2PATH_ERR(CLIENTS_NM()));
		BAD_FILE();
		fclose(f);
		return;
	}
	clients_cleared = header.cleared;
	max = min(header.hash_len, set_rl_min_max());
	rl_expand(max);

	stale = db_time.tv_sec - CLIENTS_SAVE_AGE;
	for (total = 0, used = 0; ; ++total) {
		if (!clients_read(f, &data, sizeof(data)))
			break;
		if (data.last_used > db_time.tv_sec) {
			dcc_error_msg("badly timestamped entry in %s",
				      DB_NM2PATH_ERR(CLIENTS_NM()));
			break;
		}
		if (data.last_used < stale)
			continue;	/* too stale */

		if (rl_get(&rl, data.clnt_id, &data.clnt_addr)) {
			dcc_error_msg("duplicate entry in %s",
				      DB_NM2PATH_ERR(CLIENTS_NM()));
			break;
		}

		rl->d = data;
		rl->d.flags &= ~RL_FG_CK_BL;	/* must check blacklist again */
		rl_avg_requests_age(rl, 0);
	}
	fclose(f);

	used = 0;
	for (rl = rl_oldest; rl != 0; rl = rl->newer) {
		if (rl->d.last_used != 0)
			++used;
	}
	dcc_trace_msg("used %d of %d RL blocks in %s",
		      used, total, DB_NM2PATH_ERR(CLIENTS_NM()));
}



/* pack up a 32 bit int as part of an answer to a `cdcc clients` request */
static u_char *
client_pack4(u_char *cp,
	     u_int32_t v)
{
	while (v > 0x7f) {
		*cp++ = v | 0x80;
		v >>= 7;
	}
	*cp++ = v;
	return cp;
}



static int
client_pack(u_char *cp0,
	    u_char **pflags,
#ifdef DCC_PKT_VERSION6
	    const QUEUE *q,
#endif
	     u_char flags,
	     DCC_CLNT_ID clnt_id,
	     time_t last_used,		/* skip count if *_CLIENTS_SKIP */
	     int requests,
	     int nops,
	     u_char vers,
	     const struct in6_addr *clnt_addr)
{
	u_char *cp;

#ifdef DCC_PKT_VERSION6
	if (q && q->pkt.hdr.pkt_vers <= DCC_PKT_VERSION6) {
#define CPY2(t,s) (t[0] = s>>8, t[1] = s)
#define CPY3(t,s) (t[0] = s>>16, t[1] = s>>8, t[2] = s)
#define CPY4(t,s) (t[0] = s>>24, t[1] = s>>16, t[2] = s>>8, t[3] = s)
		DCC_ADMN_RESP_CLIENTSv6 *cl;
		int addr_len;

		cl = (DCC_ADMN_RESP_CLIENTSv6 *)cp0;
		*pflags = 0;
		cl->flags = flags & (DCC_ADMN_RESP_CLIENTS_BL
				     | DCC_ADMN_RESP_CLIENTS_IPV6
				     | DCC_ADMN_RESP_CLIENTS_SKIP);
		if (flags & DCC_ADMN_RESP_CLIENTS_SKIP) {
			CPY4(cl->last_used, 0);
			CPY3(cl->requests, last_used);
		} else {
			CPY4(cl->last_used, last_used);
			CPY3(cl->requests, requests);
		}
		CPY2(cl->nops, nops);
		CPY4(cl->clnt_id, clnt_id);
		if (!clnt_addr) {
			addr_len = 4;
			memset(&cl->addr, 0, addr_len);
		} else if (DCC_IN6_ADDR_V4MAPPED(clnt_addr)) {
			addr_len = 4;
			memcpy(&cl->addr, &clnt_addr->s6_addr32[3], addr_len);
		} else {
			cl->flags |= DCC_ADMN_RESP_CLIENTS_IPV6;
			addr_len = 16;
			memcpy(&cl->addr, clnt_addr, addr_len);
		}

		return sizeof(*cl) - sizeof(cl->addr) + addr_len;
#undef CPY2
#undef CPY3
#undef CPY4
	}
#endif

	cp = cp0;
	*pflags = cp0;
	if (vers != 0)
		flags |= DCC_ADMN_RESP_CLIENTS_VERS;
	*cp++ = flags;
	if (flags & DCC_ADMN_RESP_CLIENTS_VERS)
		*cp++ = vers;
	*cp++ = last_used >> 24;
	*cp++ = last_used >> 16;
	*cp++ = last_used >> 8;
	*cp++ = last_used;
	if (clnt_id == DCC_ID_ANON) {
		*cp0 |= DCC_ADMN_RESP_CLIENTS_ID1;
	} else {
		cp = client_pack4(cp, clnt_id);
	}
	cp = client_pack4(cp, requests);
	cp = client_pack4(cp, nops);
	if (!clnt_addr) {
		memset(cp, 0, 4);
		cp += 4;
	} else if (DCC_IN6_ADDR_V4MAPPED(clnt_addr)) {
		memcpy(cp, &clnt_addr->s6_addr32[3], 4);
		cp += 4;
	} else {
		*cp0 |= DCC_ADMN_RESP_CLIENTS_IPV6;
		memcpy(cp, clnt_addr->s6_addr32, 16);
		cp += 16;
	}
	return cp-cp0;
}



static int
client_pack_skip(u_char *cp, u_char **pflags, const QUEUE *q, u_int skipped)
{
	return client_pack(cp, pflags, q, DCC_ADMN_RESP_CLIENTS_SKIP,
			   DCC_ID_ANON, skipped, 0, 0, 0, 0);
}



/* get clients, treating clients with the same ID as if they were all
 * the same system */
void
clients_get_id(DCC_ADMN_RESP_VAL *val,
	       int *lenp,		/* buffer length */
	       const QUEUE *q,
	       u_int offset,		/* skip this many newer entries */
	       int thold,		/* skip clients with fewer requests */
	       u_char req_flags,
	       const struct in6_addr *addr6,
	       const struct in6_addr *mask6)
{
	RL *rl, *rl2;
	u_char *skip;
	int requests, nops;
	u_char *pflags;
	u_int skipped;
	int len, len_lim;

	if (offset == 0)
		clients_saved = 0;

	pflags = 0;
	len_lim = *lenp;
#ifdef DCC_PKT_VERSION6
	if (q->pkt.hdr.pkt_vers <= DCC_PKT_VERSION6)
		len_lim *= sizeof(DCC_ADMN_RESP_CLIENTSv6);
	else
#endif
		len_lim <<= DCC_ADMIN_RESP_CLIENTS_SHIFT;
	if (len_lim > ISZ(*val))
		len_lim = ISZ(*val);
	skip = 0;
	skipped = 0;
	len = 0;
	if (!thold)
		thold = 1;
	for (rl = rl_newest; rl != 0; rl = rl->older)
		rl->d.flags &= ~RL_FG_MARKED;

	for (rl = rl_newest; ; rl = rl->older) {
		if (!rl) {
			if (pflags)
				*pflags |= DCC_ADMN_RESP_CLIENTS_LAST;
			break;
		}

		if (rl->d.flags & RL_FG_MARKED)
			continue;

		if (addr6 && !DCC_IN_BLOCK(rl->d.clnt_addr, *addr6, *mask6)) {
			rl->d.flags |= RL_FG_MARKED;
			continue;
		}

		if (rl->d.last_used == 0) {
			if (!(req_flags & DCC_AOP_CLIENTS_AVG))
				continue;
			/* report clients that have been quiet today if
			 * they were active yesterday */
			rl_avg_requests_age(rl, 0);
			if (RL_REQUESTS_AVG(&rl->d) == 0
			    && RL_NOPS_AVG(&rl->d) == 0)
				continue;
		}

		requests = 0;
		nops = 0;
		for (rl2 = rl; rl2 != 0; rl2 = rl2->older) {
			if (rl2->d.clnt_id != rl->d.clnt_id)
				continue;
			rl2->d.flags |= RL_FG_MARKED;
			if (addr6
			    && !DCC_IN_BLOCK(rl->d.clnt_addr, *addr6, *mask6))
				continue;
			if (req_flags & DCC_AOP_CLIENTS_AVG) {
				rl_avg_requests_age(rl2, 0);
				requests += RL_REQUESTS_AVG(&rl2->d);
				nops += RL_NOPS_AVG(&rl2->d);
			} else {
				requests += rl2->d.requests;
				nops += rl2->d.nops;
			}
		}

		/* get the part of the list that cdcc wants */
		if (offset != 0) {
			--offset;
			continue;
		}

		if (requests < thold) {
			/* The threshold might be larger on the next request
			 * from cdcc, so tell cdcc the number skipped for
			 * the threshold this time.
			 * Tell cdcc by insertint a fake entry. */
			if (!skip) {
				skip = &val->clients[len];
				len += client_pack_skip(skip, &pflags, q, 0);
			}
			++skipped;
			continue;
		}

		/* stop at end of buffer
		 * check only after skipping boring records for the common
		 * case of the last 10,000 records missing the threshold */
		if (len+DCC_ADMN_RESP_CLIENTS_MAX_SIZE > len_lim)
			break;

		len += client_pack(&val->clients[len], &pflags, q, 0,
				   rl->d.clnt_id, rl->d.last_used,
				   requests, nops, 0, 0);
	}
	if (skipped)
		client_pack_skip(skip, &pflags, q, skipped);
	*lenp = len;
}



/* get the most recent clients */
int					/* +/- number of clients */
clients_get(DCC_ADMN_RESP_VAL *val,
	    int *lenp,			/* buffer length */
#ifdef DCC_PKT_VERSION6
	    const QUEUE *q,		/* someday get rid of this parameter */
#endif
	    u_int offset,		/* skip this many newer entries */
	    int thold,			/* skip clients with fewer requests */
	    u_char req_flags,
	    const struct in6_addr *addr6,
	    const struct in6_addr *mask6)
{
	RL *rl;
	u_char *skip;
	int requests, nops;
	u_char prev_vers, vers, *pflags;
	int skipped, total, len, len_lim;

	if (offset == 0)
		clients_saved = 0;

	pflags = 0;
	if (!val || !lenp) {
		len_lim = 0;
	} else {
		len_lim = *lenp;
#ifdef DCC_PKT_VERSION6
		if (q && q->pkt.hdr.pkt_vers <= DCC_PKT_VERSION6)
			len_lim *= sizeof(DCC_ADMN_RESP_CLIENTSv6);
		else
#endif
			len_lim <<= DCC_ADMIN_RESP_CLIENTS_SHIFT;
		if (len_lim > ISZ(*val))
			len_lim = ISZ(*val);
	}
	if (!thold)
		thold = 1;
	prev_vers = 0;
	skip = 0;
	skipped = 0;
	total = 0;
	len = 0;
	for (rl = rl_newest; ; rl = rl->older) {
		if (!rl) {
			if (pflags)
				*pflags |= DCC_ADMN_RESP_CLIENTS_LAST;
			break;
		}

		if (rl->d.last_used == 0) {
			if (!(req_flags & DCC_AOP_CLIENTS_AVG))
				continue;
			/* report clients that have been quiet today if
			 * they were active yesterday */
			rl_avg_requests_age(rl, 0);
			if (RL_REQUESTS_AVG(&rl->d) == 0
			    && RL_NOPS_AVG(&rl->d) == 0)
				continue;
		}

		if (rl->d.requests != 0)
			++total;

		/* compute only the total for `cdcc stats` if buffer is null */
		if (len_lim == 0)
			continue;

		/* respect client IP address limitations */
		if (addr6 && !DCC_IN_BLOCK(rl->d.clnt_addr, *addr6, *mask6))
			continue;

		/* report only (non-)anonymous clients if asked */
		if ((req_flags & DCC_AOP_CLIENTS_ANON)
		    && rl->d.clnt_id != DCC_ID_ANON)
			continue;
		if ((req_flags & DCC_AOP_CLIENTS_NON_ANON)
		    && rl->d.clnt_id == DCC_ID_ANON)
			continue;

		if (offset != 0) {
			--offset;
			continue;
		}

		if (req_flags & DCC_AOP_CLIENTS_AVG) {
			rl_avg_requests_age(rl, 0);
			requests = RL_REQUESTS_AVG(&rl->d);
			nops = RL_NOPS_AVG(&rl->d);
		} else {
			requests = rl->d.requests;
			nops = rl->d.nops;
		}

		/* skip uninteresting records and insert a fake
		 * entry in the output to the client with the total
		 * skipped in the entire response to the client */
		if (requests < thold
		    && (!(rl->d.flags & (RL_FG_ADDR_BL | RL_FG_ID_BL))
			|| rl->d.requests == 0)) {
			if (!skip) {
				/* start a new fake record */
				skip = &val->clients[len];
				len += client_pack_skip(skip, &pflags, q, 0);
				prev_vers = 0;
			}
			++skipped;
			continue;
		}

		/* stop at end of buffer
		 * check only after skipping boring records for the common
		 * case of ignoring the last 10,000 records */
		if (len + DCC_ADMN_RESP_CLIENTS_MAX_SIZE*2 > len_lim)
				break;

		/* send the version number if it is wanted and differs
		 * from the previous value */
		if ((req_flags & DCC_AOP_CLIENTS_VERS)
		    && rl->d.pkt_vers != prev_vers) {
			vers = rl->d.pkt_vers;
			prev_vers = vers;
		} else {
			vers = 0;
		}

		len += client_pack(&val->clients[len], &pflags, q,
				   (rl->d.flags & (RL_FG_ADDR_BL | RL_FG_ID_BL))
				   ? DCC_ADMN_RESP_CLIENTS_BL : 0,
				   rl->d.clnt_id, rl->d.last_used,
				   requests, nops, vers, &rl->d.clnt_addr);
	}
	/* put final total number of skipped records in the output */
	if (skipped)
		client_pack_skip(skip, &pflags, q, skipped);

	if (lenp)
		*lenp = len;

	/* return negative total if we have more clients than RL blocks */
	if (rl_too_many)
		return -total;
	return total;
}



/* forget old clients */
void
clients_clear(u_char force_rl_age)
{
	RL *rl;
	;
	for (rl = rl_oldest; rl != 0; rl = rl->newer) {
		rl->d.requests_next_avg += rl->d.requests;
		rl->d.requests = 0;
		rl->d.nops_next_avg += rl->d.nops;
		rl->d.nops = 0;
		rl_avg_requests_age(rl, force_rl_age);
	}

	clients_cleared = db_time.tv_sec;
	rl_too_many = 0;
}



u_char					/* 0=bad passwd, 1=1st passwd, 2=2nd */
ck_sign(const ID_TBL **tpp,		/* return ID table entry here */
	DCC_PASSWD passwd,		/* return matching password here */
	DCC_CLNT_ID id,
	const void *buf, u_int buf_len)
{
	ID_TBL *tp;

	tp = find_id_tbl(id);
	if (tpp)
		*tpp = tp;
	if (!tp)
		return 0;

	if (tp->cur_passwd[0] != '\0'
	    && dcc_ck_signature(tp->cur_passwd, sizeof(tp->cur_passwd),
				buf, buf_len)) {
		if (passwd)
			memcpy(passwd, tp->cur_passwd, sizeof(DCC_PASSWD));
		return 1;
	}
	if (tp->next_passwd[0] != '\0'
	    && dcc_ck_signature(tp->next_passwd, sizeof(tp->next_passwd),
				buf, buf_len)) {
		if (passwd)
			memcpy(passwd, tp->next_passwd, sizeof(DCC_PASSWD));
		return 2;
	}
	return 0;
}



/* decide how long a request should be delayed */
static void
tp2delay(QUEUE *q, time_t delay_us, u_int delay_inflate)
{
	int inflate;

	if (delay_us == 0) {
		q->delay_us = 0;
	} else {
		/* do not answer anonymous clients while we are cleaning
		 * the database */
		if (db_minimum_map && q->clnt_id == DCC_ID_ANON) {
			q->delay_us = DCC_ANON_DELAY_MAX;
			return;
		}
		inflate = 1 + RL_REQUESTS_AVG(&q->rl->d)/delay_inflate;
		if (inflate > DCC_ANON_DELAY_MAX / delay_us) {
			/* prevent overflow */
			q->delay_us = DCC_ANON_DELAY_MAX;
			return;
		}
		q->delay_us = inflate * delay_us;
	}

	if (flods_off > 0
	    || flods_st != FLODS_ST_ON
	    || (iflods.open == 0 && (oflods.total != 0 || !grey_on))
	    || !DB_IS_TIME(iflods_ok_timer, IFLODS_OK_SECS)) {
		/* Increase the delay when flooding is off, broken,
		 * or just starting after being off
		 * Greylist servers without any flooding peers can be isolated
		 * but not DCC servers */
		q->delay_us += 400*1000;

	} else if (dbclean_wfix_state == WFIX_QUIET
		   || dbclean_wfix_state == WFIX_CHECK) {
		/* increase the delay if we are testing to see if clients
		 * have a alternative so we can do a quick window fix. */
		q->delay_us += 1000*1000;
	}

	if (q->delay_us > DCC_ANON_DELAY_MAX)
		q->delay_us = DCC_ANON_DELAY_MAX;
}



static const char *
ck_id_str(const char *bad_id_str, DCC_CLNT_ID id)
{
	static char ck_id_str_buf[sizeof("unauthenticated client-ID xxxxxxxx")];

	if (id == DCC_ID_ANON)
		return "anonymous";

	snprintf(ck_id_str_buf, sizeof(ck_id_str_buf), "%sclient-ID %d",
		 bad_id_str, id);
	return ck_id_str_buf;
}



/* check the message authentication code and rate limit requests */
u_char					/* 0=forget DOS attack, 1=go ahead */
ck_id(QUEUE *q, DCC_CLNT_ID id)
{
#define RL_CNT2AVG(cur,lims) ((lims.hi - cur) / (RL_SCALE*RL_AVG_SECS*1.0))
	const ID_TBL *tp;
	RL *rl;
	const char *bad_id_str;
	char id_bl;

	id_bl = 0;
	bad_id_str = "";
	if (id == DCC_ID_ANON) {
		tp = 0;
	} else {
		/* authenticate the ID */
		if (!ck_sign(&tp, q->passwd, id, &q->pkt, q->pkt_len)) {
			if (!tp) {
				bad_id_str = "unknown ";
				++dccd_stats.unknown_ids;
			} else {
				if (tp->delay_us == DCC_ANON_DELAY_US_BLACKLIST)
					id_bl = 1;
				bad_id_str = "unauthenticated ";
				tp = 0;
				++dccd_stats.bad_passwd;
			}
		} else if (tp->delay_us == DCC_ANON_DELAY_US_BLACKLIST) {
			id_bl = 1;
		}
	}

	q->clnt_id = (tp || id_bl) ? id : DCC_ID_ANON;
	rl = rl_get_q(q);
	rl->d.pkt_vers = q->pkt.hdr.pkt_vers;
	if (id_bl)
		rl->d.flags |= RL_FG_ID_BL;
	else
		rl->d.flags &= ~RL_FG_ID_BL;

	if (((rl->d.flags & (RL_FG_ADDR_BL | RL_FG_ID_BL))
	     && (DCC_TRACE_BL_BIT & dccd_tracemask))
	    || (rl->d.flags & RL_FG_TRACE)
	    || (*bad_id_str != '\0' && !(RL_FG_ADDR_BL | RL_FG_ID_BL)))
		dcc_trace_msg("%s at %s%s %s",
			      ck_id_str(bad_id_str, id),
			      Q_CIP(q),
			      (rl->d.flags & RL_FG_ADDR_BL) ? " blacklisted"
			      : (rl->d.flags & RL_FG_ID_BL) ? " ID blacklistd"
			      : "",
			      dcc_hdr_op2str(0, 0, &q->pkt.hdr));

	if (tp) {
		if (!query_only || (tp->flags & ID_FLG_RPT_OK))
			q->flags |= Q_FLG_RPT_OK;

		rl_inc(rl, &rl_sub_rate);

		if (rl->d.flags & (RL_FG_ADDR_BL | RL_FG_ID_BL))
			return 0;

		if (rl->d.request_credits <= 0) {
			clnt_msg(q, "%.1f requests/sec are too many from %s %s",
				 RL_CNT2AVG(rl->d.request_credits, rl_sub_rate),
				 ck_id_str(bad_id_str, id),
				 Q_CIP(q));
			++dccd_stats.rl;
			return 0;
		}

		tp2delay(q, tp->delay_us, tp->delay_inflate);

	} else {
		if (!query_only)
			q->flags |= Q_FLG_RPT_OK;

		rl_inc(rl, &rl_anon_rate);
		rl_inc(&rl_all_anon, &rl_all_anon_rate);

		if (rl->d.flags & (RL_FG_ADDR_BL | RL_FG_ID_BL))
			return 0;

		if (rl->d.request_credits <= 0) {
			anon_msg("%.1f requests/sec are too many from %s %s",
				 RL_CNT2AVG(rl->d.request_credits,
					    rl_anon_rate),
				 ck_id_str(bad_id_str, id),
				 Q_CIP(q));
			++dccd_stats.anon_rl;
			return 0;
		}
		if (rl_all_anon.d.request_credits <= 0) {
			anon_msg("%s contributed to %.1f"
				 " anonymous requests/sec",
				 Q_CIP(q),
				 RL_CNT2AVG(rl_all_anon.d.request_credits,
					    rl_all_anon_rate));
			++dccd_stats.anon_rl;
			return 0;
		}

		tp2delay(q, anon_delay_us, anon_delay_inflate);
	}

	return 1;
}



/* check the message authentication code for a client of our server-ID
 * and rate limit its messages */
u_char					/* 0=forget DOS attack, 1=go ahead */
ck_clnt_srvr_id(QUEUE *q)
{
	DCC_CLNT_ID id;
	u_char result;
	char ob[DCC_OPBUF];

	/* require a client-ID, our server-ID, or the anonymous client-ID
	 * to consider allowing an administrative request */
	id = ntohl(q->pkt.hdr.sender);
	if ((id < DCC_SRVR_ID_MIN || id > DCC_CLNT_ID_MAX)
	    && id != DCC_ID_ANON) {
		++dccd_stats.unknown_ids;
		result = ck_id(q, DCC_ID_ANON);
		if (result)
			anon_msg("bad client or server-ID %d from %s for %s",
				 id, Q_CIP(q),
				 dcc_hdr_op2str(ob, sizeof(ob), &q->pkt.hdr));
		return result;
	}
	return ck_id(q, id);
}



/* check the message authentication code of a request,
 * and rate limit the source */
u_char					/* 0=forget DOS attack, 1=go ahead */
ck_clnt_id(QUEUE *q)
{
	DCC_CLNT_ID id;
	u_char result;
	char ob[DCC_OPBUF];

	/* require a client-ID instead of a server-ID to discourage server
	 * operators from leaking server-ID's */
	id = ntohl(q->pkt.hdr.sender);
	if (id < DCC_CLNT_ID_MIN
	    && id != DCC_ID_ANON) {
		++dccd_stats.unknown_ids;
		result = ck_id(q, DCC_ID_ANON);
		if (result)
			anon_msg("bad client-ID %d from %s for %s",
				 id, Q_CIP(q),
				 dcc_hdr_op2str(ob, sizeof(ob), &q->pkt.hdr));
		return result;
	}
	return ck_id(q, id);
}



/* complain about an anonymous, non-paying client */
void
vanon_msg(const char *p, va_list args)
{
	rl_age(&rl_all_anon, &rl_all_anon_rate);
	if ((DCC_TRACE_ANON_BIT & dccd_tracemask)
	    && (rl_all_anon.d.bug_credits > 0
		|| (DCC_TRACE_RLIM_BIT & dccd_tracemask))) {
		rl_all_anon.d.bug_credits -= RL_SCALE;
		dcc_vtrace_msg(p, args);
	}
}



void
anon_msg(const char *p, ...)
{
	va_list args;

	va_start(args, p);
	vanon_msg(p, args);
	va_end(args);
}



/* complain about a client */
void
vclnt_msg(const QUEUE *q, const char *p, va_list args)
{
	if (q->clnt_id == DCC_ID_ANON || !q->rl) {
		vanon_msg(p, args);
		return;
	}

	if (DCC_TRACE_CLNT_BIT & dccd_tracemask) {
		rl_age(q->rl, &rl_sub_rate);
		if (q->rl->d.bug_credits > 0
		    || (DCC_TRACE_RLIM_BIT & dccd_tracemask)) {
			q->rl->d.bug_credits -= RL_SCALE;
			dcc_vtrace_msg(p, args);
		}
	}
}



/* complain about a client */
void
clnt_msg(const QUEUE *q, const char *p, ...)
{
	va_list args;

	va_start(args, p);
	vclnt_msg(q, p, args);
	va_end(args);
}
